@@ -5,6 +5,7 @@ from __future__ import absolute_import, unicode_literals
 import sys
 import copy
 from functools import partial
+import warnings
 
 try:
     str = unicode  # Python 2.* compatible
@@ -23,6 +24,9 @@ class SqlDialects(object):
     """
     Stores all dialect representations
     """
+
+    __slots__ = ('_registry', )
+
     def __init__(self):
         """Constructor, initial registry."""
         self._registry = {}
@@ -82,8 +86,7 @@ class Error(Exception):
 
 class Expr(object):
 
-    _sql = None
-    _params = None
+    __slots__ = ('_sql', '_params')
 
     def __init__(self, sql, *params):
         self._sql = sql
@@ -245,10 +248,10 @@ class Expr(object):
             return self.__eq__(key)
 
     def __sqlrepr__(self, dialect):
-        return self._sql or ""
+        return getattr(self, '_sql', "")
 
     def __params__(self):
-        return self._params or []
+        return getattr(self, '_params', [])
 
     def __bytes__(self):
         return sqlrepr(self).encode('utf-8')
@@ -270,6 +273,9 @@ class Expr(object):
 
 
 class Condition(Expr):
+
+    __slots__ = ('_op', '_expr1', '_expr2')
+
     def __init__(self, op, expr1, expr2):
         self._op = op.upper()
         if expr1 is not None:
@@ -295,6 +301,8 @@ class Condition(Expr):
 
 class ExprList(Expr):
 
+    __slots__ = ('_sep', '_args')
+
     def __init__(self, *args):
         self._sep = " "
         self._args = []
@@ -361,8 +369,16 @@ class ExprList(Expr):
             params.extend(sqlparams(arg))
         return params
 
+    def __copy__(self):
+        dup = copy.copy(super(ExprList, self))
+        dup._args = dup._args[:]
+        return dup
+
 
 class Concat(ExprList):
+
+    __slots__ = ('_sep', '_args', '_ws')
+
     def __init__(self, *args):
         super(Concat, self).__init__(*args)
         self._sep = ' || '
@@ -386,12 +402,17 @@ class Concat(ExprList):
 
 
 class Placeholder(Expr):
+
+    __slots__ = ('_sql', '_params')
+
     def __init__(self, *params):
         super(Placeholder, self).__init__(PLACEHOLDER, *params)
 
 
 class Parentheses(Expr):
 
+    __slots__ = ('_expr', )
+
     def __init__(self, expr):
         self._expr = expr
 
@@ -404,6 +425,8 @@ class Parentheses(Expr):
 
 class Prefix(Expr):
 
+    __slots__ = ('_prefix', '_expr', )
+
     def __init__(self, prefix, expr):
         self._prefix = prefix
         self._expr = prepare_expr(expr)
@@ -417,6 +440,8 @@ class Prefix(Expr):
 
 class Between(Expr):
 
+    __slots__ = ('_expr', '_start', '_end')
+
     def __init__(self, expr, start, end):
         self._expr = prepare_expr(expr)
         self._start = prepare_expr(start)
@@ -436,6 +461,8 @@ class Between(Expr):
 
 class Callable(Expr):
 
+    __slots__ = ('_expr', '_args')
+
     def __init__(self, expr, *args):
         self._expr = expr
         self._args = ExprList(*args).join(", ")
@@ -448,6 +475,9 @@ class Callable(Expr):
 
 
 class Constant(Expr):
+
+    __slots__ = ('_const', '_params')
+
     def __init__(self, const):
         self._const = const.upper()
         self._params = []
@@ -460,7 +490,10 @@ class Constant(Expr):
         return self._const
 
 
-class ConstantSpace:
+class ConstantSpace(object):
+
+    __slots__ = ()
+
     def __getattr__(self, attr):
         if attr.startswith('__'):
             raise AttributeError
@@ -489,6 +522,9 @@ class MetaField(type):
 
 
 class Field(MetaField(bytes("NewBase"), (Expr, ), {})):
+
+    __slots__ = ('_name', )
+
     def __init__(self, name, prefix=None):
         self._name = name
 
@@ -504,6 +540,9 @@ class Field(MetaField(bytes("NewBase"), (Expr, ), {})):
 
 
 class Alias(Expr):
+
+    __slots__ = ('_expr', )
+
     def __init__(self, alias, expr=None):
         self._expr = expr
         super(Alias, self).__init__(alias)
@@ -537,6 +576,9 @@ class MetaTable(type):
 
 
 class Table(MetaTable(bytes("NewBase"), (object, ), {})):
+
+    __slots__ = ('_name', )
+
     def __init__(self, name):
         self._name = name
 
@@ -616,6 +658,9 @@ class Table(MetaTable(bytes("NewBase"), (object, ), {})):
 
 
 class TableAlias(Table):
+
+    __slots__ = ('_table', '_alias')
+
     def __init__(self, alias, table):
         self._table = table
         self._alias = alias
@@ -633,6 +678,8 @@ class TableAlias(Table):
 
 class TableJoin(object):
 
+    __slots__ = ('_table', '_alias', '_join_type', '_on', '_left', '_use_index', '_ignore_index', '_force_index', )
+
     def __init__(self, table_or_alias, join_type=None, on=None, left=None):
         if isinstance(table_or_alias, TableAlias):
             self._table = table_or_alias.table
@@ -687,9 +734,6 @@ class TableJoin(object):
     def group(self):
         return TableJoin(self)
 
-    def as_nested(self):
-        return self.group()
-
     @opt_checker(["reset", ])
     def change_index(self, index, *args, **opts):
         if opts.get("reset"):
@@ -718,6 +762,12 @@ class TableJoin(object):
     def force_index(self, *args, **opts):
         return self.change_index(self._force_index, *args, **opts)
 
+    def __copy__(self):
+        dup = copy.copy(super(TableJoin, self))
+        for a in ['_use_index', '_ignore_index', '_force_index', ]:
+            setattr(dup, a, copy.copy(getattr(dup, a, None)))
+        return dup
+
     def __sqlrepr__(self, dialect):
         sql = ExprList().join(" ")
         if self._left is not None:
@@ -754,6 +804,7 @@ class TableJoin(object):
         return sqlrepr(self)
 
     # Aliases:
+    as_nested = same('group')
     __and__ = same('inner_join')
     __add__ = same('left_join')
     __sub__ = same('right_join')
@@ -809,7 +860,11 @@ class QuerySet(Expr):
         self._havings = cs
 
     def clone(self):
-        return copy.deepcopy(self)
+        # return copy.deepcopy(self)
+        dup = copy.copy(super(QuerySet, self))
+        for a in ['_fields', '_tables', '_group_by', '_order_by', '_values', '_key_values', ]:
+            setattr(dup, a, copy.copy(getattr(dup, a, None)))
+        return dup
 
     def dialect(self, dialect=None):
         if dialect is not None:
@@ -861,7 +916,7 @@ class QuerySet(Expr):
         self = self.clone()
         if not isinstance(self._tables, TableJoin):
             raise Error("Can't set on without join table")
-        self._tables.on(c)
+        self._tables = self._tables.on(c)
         return self
 
     def where(self, c):
@@ -1144,6 +1199,7 @@ class QuerySet(Expr):
 
     # Aliases:
     columns = same('fields')
+    __copy__ = same('clone')
 
 
 class UnionQuerySet(QuerySet):
@@ -1170,8 +1226,16 @@ class UnionQuerySet(QuerySet):
         self._sql_extend(sql, ["order", "limit"])
         return sql
 
+    def clone(self):
+        self = super(UnionQuerySet, self).clone()
+        self._union_list = copy.copy(self._union_list)
+        return self
+
 
 class Name(object):
+
+    __slots = ('_name', )
+
     def __init__(self, name=None):
         self._name = name
 
@@ -1235,6 +1299,14 @@ def sqlparams(obj):
         return list(obj.__params__())
     return []
 
+
+def warn(old, new, stacklevel=3):
+    warnings.warn(
+        "{0} is deprecated. Use {1} instead".format(old, new),
+        PendingDeprecationWarning,
+        stacklevel=stacklevel
+    )
+
 T, TA, F, A, E, QS = Table, TableAlias, Field, Alias, Expr, QuerySet
 const = ConstantSpace()
 func = const
